# (c) 2017-2019, Tal Shany <tal.shany@biSkilled.com>
#
# This file is part of dingDong
#
# dingDong is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# dingDong is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with dingDong.  If not, see <http://www.gnu.org/licenses/>.

import re
import sys

from dingDong.config import config

def replaceStr (sString,findStr, repStr, ignoreCase=True,addQuotes=None):
    if addQuotes and isinstance(repStr,str):
        repStr="%s%s%s" %(addQuotes,repStr,addQuotes)

    if ignoreCase:
        pattern = re.compile(re.escape(findStr), re.IGNORECASE)
        res = pattern.sub (repStr, sString)
    else:
        res = sString.replace (findStr, repStr)
    return res

def uniocdeStr (sObj, decode=False):
    if 3 == sys.version_info[0]:
        return sObj

    return unicode (str(sObj).decode(config.DECODE)) if decode else unicode (sObj)



# (c) 2017-2019, Tal Shany <tal.shany@biSkilled.com>
#
# This file is part of dingDong
#
# dingDong is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# dingDong is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with dingDong.  If not, see <http://www.gnu.org/licenses/>.

from collections import OrderedDict

class eJson (object):

    class jKeys (object):
        #  JSON TYPES Keys
        TARGET      = 't'
        SOURCE      = 's'
        MERGE       = 'm'
        QUERY       = 'q'
        SEQUENCE    = 'seq'
        STT         = 'sttappend'
        STTONLY     = 'stt'
        MAP         = 'map'
        COLUMNS     = 'col'
        INDEX       = 'index'
        PARTITION   = 'par'           # not inplemented
        INC         = 'inc'           # not implemented
        NONO        = 'internal'

        eDict = {
            TARGET: [TARGET, 'target'],  # Target
            SOURCE: [SOURCE, 'source'],  # Source
            QUERY: [QUERY, 'query'],
            MERGE: [MERGE, 'merge'],
            SEQUENCE: [SEQUENCE, 'sequence'],
            STT: [STT, 'sourcetotarget'],
            STTONLY: [STTONLY, 'only'],
            MAP: [MAP, 'mapping'],
            COLUMNS: [COLUMNS, 'columns', 'column'],
            INDEX: [INDEX, 'i'],
            PARTITION: [PARTITION, 'partition'],
            INC: [INC, 'incremental']
        }

    class jValues(object):
        # Internal Json keys {'s':{'conn'}}
        NAME        = 'n'
        TYPE        = 'type'
        CONN        = 'conn'
        OBJ         = 'obj'
        FILTER      = 'filter'
        IS_CHANGE   = 'ischange'
        URL         = 'url'
        URLPARAM    = 'eurl'
        URL_FILE    = 'urlfile'
        DEFAULT_TYPE= 'defaultdatatype'
        SCHEMA      = 'schema'
        EMPTY       = 'empty'
        COLFRAME    = 'colFrame'
        SP          = 'sp'
        IS_SOURCE   = 'src'
        IS_TARGET   = 'tar'
        IS_SQL      = 'sql'
        FOLDER      = 'folder'
        BATCH_SIZE  = 'batch'
        UPDATE      = 'update'
        INDEX_COLUMS  = 'c'
        INDEX_CLUSTER = 'ic'
        INDEX_UNIQUE  = 'iu'


        eDict = {
            NAME:       [NAME, 'name'],
            TYPE:       [TYPE],
            OBJ:        [OBJ, 'object'],
            FILTER:     [FILTER, 'filter'],
            IS_CHANGE:  [IS_CHANGE, 'model', 'isCahnged'],
            URL:        [URL],
            URLPARAM:   [URLPARAM, 'extra', 'extraparam'],
            FOLDER:     [FOLDER, 'files'],
            BATCH_SIZE: [BATCH_SIZE,'totalrows', 'rows'],
            URL_FILE  : [URL_FILE, 'file'],
            UPDATE:     [UPDATE,'change'],
            INDEX_CLUSTER: [INDEX_CLUSTER, 'cluster'],
            INDEX_UNIQUE:[INDEX_UNIQUE, 'unique'],
            INDEX_COLUMS:[INDEX_COLUMS,'column','columns']
        }
        DIC   = {   NAME:None,TYPE:None,CONN:None,OBJ:None,
                    FILTER:None,IS_CHANGE:None,URL:None,URLPARAM:None,
                    URL_FILE:None,DEFAULT_TYPE:None,SCHEMA:None,EMPTY:None,COLFRAME:None,SP:None,
                    IS_SOURCE:None,IS_TARGET:None, IS_SQL:False, FOLDER:None}

    class jSttValues(object):
        SOURCE  = 's'
        TYPE    = 't'
        ALIACE  = 'a'
        FUNCTION= 'f'
        EXECFUNC= 'e'
        INDEX   = 'i'
        DIC     = {SOURCE: None, TYPE: None, ALIACE:None}

    class jMergeValues(object):
        SOURCE  = 'merge_source'
        TARGET  = 'merge_target'
        MERGE   = 'merge_keys'
        DIC     = {SOURCE:None, TARGET:None, MERGE:None}

    class jFile(object):
        MIN_SIZE        = 'min'
        DEF_COLUMN_PREF = 'defCol'
        ENCODING        = 'encoding'
        DELIMITER       = 'delimiter'
        ROW_HEADER      = 'headerline'
        END_OF_LINE     = "eol"
        MAX_LINES_PARSE = 'maxlines'
        LOAD_WITH_CHAR_ERR = 'charerror'
        APPEND          = 'append'
        REPLACE_TO_NONE = 'replace'
        CSV             = 'csv'

    class jUpdate(object):
        DROP      = -1
        UPDATE    = 1
        NO_UPDATE = 2

class eConn (object):
    SQLSERVER   = "sql"
    ORACLE      = "oracle"
    VERTICA     = "vertica"
    ACCESS      = "access"
    MYSQL       = "mysql"
    LITE        = "sqlite"
    FILE        = "file"
    NONO        = 'nono'


    class dataType(object):
        B_STR       = 'str'
        B_INT       = 'int'
        B_FLOAT     = 'float'
        B_LONG_STR  = 'text'
        B_BLOB      = 'blob'
        B_DEFAULT   = 'default'

        DB_VARCHAR  = 'varchar'
        DB_VARCHAR_MAX='varchar(MAX)'
        DB_NVARCHAR = 'nvarchar'
        DB_CHAR     = 'char'
        DB_BLOB     = 'blob'
        DB_INT      = 'int'
        DB_BIGINT   = 'bigint'
        DB_FLOAT    = 'float'
        DB_NUMERIC = 'numeric'
        DB_DECIMAL = 'decimal'
        DB_DATE     = 'datetime1'

class eSql (object):
    RENAME      = 'rename'
    DROP        = 'drop'
    TRUNCATE    = 'truncate'
    STRUCTURE   = 'structure'
    MERGE       = 'merge'
    ISEXISTS    = 'isexists'
    DELETE      = 'delete'
    INDEX       = 'index'
    INDEX_EXISTS= 'indexexists'

    TABLE_COPY_BY_COLUMN = 'copy'

""" Methods To use for fun """
def findProp (prop, obj=eConn, dictProp=None):
    dicClass = obj.__dict__
    def getPropValue (prop):
        if prop:
            for p in dicClass:
                if isinstance(dicClass[p], str) and dicClass[p].lower() == str(prop).lower():
                    return prop
                elif isinstance(dicClass[p], int) and str(dicClass[p]) == str(prop):
                    return prop

                if isinstance(dicClass[p], dict):
                    for k in dicClass[p]:
                        if dicClass[p][k] and prop in dicClass[p][k]:
                            return k
        return None

    ret = getPropValue (prop)
    if ret is not None:
        return ret
    if ret is None and dictProp and isinstance(dictProp ,(dict,OrderedDict)):
        keys = list(dictProp.keys())
        for k in keys:
            ret = getPropValue(k)
            if ret is not None:
                return ret
    return None


def getAllProp (obj=eJson):
    dicClass = obj.__dict__
    ret = []
    for p in dicClass:
        if '__' not in p and isinstance(dicClass[p], str):
            ret.append ( dicClass[p] )
    return ret



# -*- coding: utf-8 -*-
# (c) 2017-2019, Tal Shany <tal.shany@biSkilled.com>
#
# This file is part of dingDong
#
# dingDong is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# dingDong is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with dingDong.  If not, see <http://www.gnu.org/licenses/>.

from dingDong.misc.enumsJson import getAllProp, eConn, eSql
from dingDong.misc.logger import p

""" Abstract Class: """
class baseSqlQuery (object):
    def __init__ (self):
        self.allConn    = getAllProp (obj=eConn)
        self.connQuery  = {}
        self.default    = None

    def initSqlDict (self, defSql=None, conn=None):
        defSql = defSql if defSql else self.default
        connList = [conn] if conn else list (self.connQuery.keys())
        for conn in connList:
            self.connQuery[conn] = defSql

    """ Return SQL Query by connection type (sql, oracle ...) and sqlType string:
        RENAME , DROP, TRUNCATE, STRUCTURE, MERGE, ISEXISTS     
    """
    def getSql(self, conn, sqlType, **args):
        self.default = None
        for c in self.allConn:
            self.connQuery[c] = None

        if eSql.RENAME      == sqlType: self.setSqlRename(**args)
        elif eSql.DROP      == sqlType: self.setSqlDrop(**args)
        elif eSql.TRUNCATE  == sqlType: self.setSqlTruncate(**args)
        elif eSql.STRUCTURE == sqlType: self.setSqlTableStructure(**args)
        elif eSql.MERGE     == sqlType: self.setSqlMerge(**args)
        elif eSql.ISEXISTS  == sqlType: self.setSqlIsExists(**args)
        elif eSql.DELETE    == sqlType: self.setSqlDelete(**args)
        elif eSql.TABLE_COPY_BY_COLUMN    == sqlType: self.tblCopyByColumn(**args)
        elif eSql.INDEX_EXISTS == sqlType: self.setSqlExistingIndexes(**args)
        elif eSql.INDEX     == sqlType: self.setSqlIndex(**args)

        else:
            p("baseConnDbSqlQueries->getSql: %s IS NOT DEFINED  !" % (sqlType.upper()), "e")
            return None

        if conn not in self.connQuery:
            p("baseConnDbSqlQueries->getSqlRename: %s SQL QUERY FOR CONNENTION %s NOT IMPLEMENTED !" % (sqlType.upper(),conn), "e")
            return None

        if not self.connQuery[conn]:
            p("baseConnDbSqlQueries->getSqlRename: %s SQL QUERY FOR CONNECTION %s USING DEFAULT SQL " % (sqlType.upper(),conn),"ii")
            return self.default

        return self.connQuery[conn]

    def setSqlRename(self, tableSchema, tableName, tableNewName):
        pass

    def setSqlDrop (self, tableName, tableSchema):
        pass

    def setSqlTruncate (self, tableName, tableSchema):
        pass

    def setSqlTableStructure(self, tableName, tableSchema):
        pass

    def setSqlMerge (self, dstTable, srcTable, mergeKeys, colList , colFullList):
        pass

    def setSqlIsExists(self, tableName, tableSchema):
        pass

    def setSqlDelete (self, sqlFilter, tableName, tableSchema):
        pass

    def tblCopyByColumn(self, tableName, tableSchema, srcTableName, columns):
        pass

    def setSqlExistingIndexes(self, tableName):
        pass

    def setSqlIndex(self,  tableName, columns, isCluster, isUnique):
        pass




class setSqlQuery (baseSqlQuery):
    def __init__ (self):
        baseSqlQuery.__init__(self)

    def setSqlRename (self, tableSchema, tableName, tableNewName):
        tableName = '%s.%s' %(tableSchema, tableName) if tableSchema else tableName
        self.default = "ALTER TABLE %s RENAME TO %s;" %(tableName, tableNewName)
        self.connQuery[eConn.NONO] = self.default
        self.connQuery[eConn.LITE] = self.default

        self.connQuery[eConn.SQLSERVER] = "EXEC sp_rename '%s','%s'" %(tableName, tableNewName)


    def setSqlDrop (self, tableName, tableSchema):
        tableName = '%s.%s' % (tableSchema, tableName) if tableSchema else tableName
        self.default = "drop table %s " %(tableName)
        self.connQuery[eConn.SQLSERVER] = self.default
        self.connQuery[eConn.ORACLE]    = self.default
        self.connQuery[eConn.LITE] = self.default

    def setSqlTruncate (self, tableName, tableSchema):
        fullTableName = '%s.%s' % (tableSchema, tableName) if tableSchema else tableName
        self.default = "truncate table %s;" %(fullTableName)
        self.connQuery[eConn.SQLSERVER] = self.default
        self.connQuery[eConn.ORACLE]    = self.default
        self.connQuery[eConn.LITE] = "DELETE FROM %s;" %(fullTableName)

    def setSqlTableStructure (self, tableName, tableSchema):

        self.default = "SQL Structure is not implemented ;"
        #### SQL SERVER
        sql = """
        SELECT c.name,
        UPPER(tp.name) +
        CASE	WHEN tp.name IN ('varchar', 'char', 'varbinary', 'binary', 'text')
                    THEN '(' + CASE WHEN c.max_length = -1 THEN 'MAX' ELSE CAST(c.max_length AS VARCHAR(5)) END + ')'
                WHEN tp.name IN ('nvarchar', 'nchar')
                    THEN '(' + CASE WHEN c.max_length = -1 THEN 'MAX' ELSE CAST(c.max_length / 2 AS VARCHAR(5)) END + ')'
                WHEN tp.name IN ('ntext')
                    THEN ''
                WHEN tp.name IN ('datetime2', 'time2', 'datetimeoffset')
                    THEN '(' + CAST(c.scale AS VARCHAR(5)) + ')'
                WHEN tp.name IN ('decimal','numeric')
                    THEN '(' + CAST(c.[precision] AS VARCHAR(5)) + ',' + CAST(c.scale AS VARCHAR(5)) + ')'
            ELSE ''
            END as colType

        FROM sys.columns c WITH (NOWAIT)
        JOIN sys.types tp WITH (NOWAIT) ON c.user_type_id = tp.user_type_id
        WHERE c.[object_id] =
            (Select top 1 object_id as obID from
            (SELECT SCHEMA_NAME(schema_id) schemaDesc , name , object_id FROM sys.tables
             Union
            Select SCHEMA_NAME(schema_id) schemaDesc , name , object_id FROM sys.views) tt
            Where """
        sql += "schemaDesc='" + tableSchema.replace("[", "").replace("]", "") + "' and " if tableSchema else ""
        sql += "name='" + tableName.replace("[", "").replace("]", "") + "') ORDER BY c.column_id"

        self.connQuery[eConn.SQLSERVER] = sql
        self.connQuery[eConn.LITE]      = "select name, type from pragma_table_info('%s');" %tableName

        ### ORACLE
        sql = "select column_name, data_type || "
        sql += "case when data_type = 'NUMBER' and data_precision is not null then '(' | | data_precision | | ',' | | data_scale | | ')' "
        sql += "when data_type = 'NUMBER' then '(18,0)' "
        sql += "when data_type like '%CHAR%' then '(' | | data_length | | ')' "
        sql += "else '' end type "
        sql += "from all_tab_columns where table_name = '" + tableName + "' "
        sql += " and owner='" + tableSchema + "'" if tableSchema else ""
        sql += " ORDER BY COLUMN_ID"

        self.connQuery[eConn.ORACLE] = str(sql)

        ### MYSQL
        sql = """
        SELECT distinct column_name, column_type  FROM information_schema.columns
        WHERE table_name='""" + tableName + "' "
        sql += "and TABLE_SCHEMA='" + tableSchema + "';" if tableSchema else ";"

        self.connQuery[eConn.MYSQL] = str(sql)

        ### VERTICA
        sql = """
            SELECT distinct column_name, data_type  FROM columns
            WHERE table_name='""" + tableName + "' "
        sql += "and table_schema='" + tableSchema + "';" if tableSchema else ";"

        self.connQuery[eConn.VERTICA] = str(sql)

    def setSqlMerge (self, dstTable, srcTable, mergeKeys, colList , colFullList):
        ### SQL AND DEFAULT
        sql = "MERGE INTO " + dstTable + " as t USING " + srcTable + " as s ON ("
        colOnMerge = " AND ".join(["ISNULL (t." + c + ",'')= ISNULL (s." + c + ",'')" for c in mergeKeys])
        sql += colOnMerge + ") \n WHEN MATCHED THEN UPDATE SET \n"
        for c in colList:
            # Merge only is source is not null
            sql += "t." + c + "=" + "case when s." + c + " is null or len(s." + c + ")<1 then t." + c + " else s." + c + " End,\n"
        sql = sql[:-2] + "\n"
        sql += " WHEN NOT MATCHED THEN \n"
        sql += " INSERT (" + ",".join([c for c in colFullList]) + ") \n"
        sql += " VALUES  (" + ",".join(["s." + c for c in colFullList]) + "); "

        self.default = sql
        self.connQuery[eConn.SQLSERVER] = sql

    def setSqlIsExists(self, tableName, tableSchema):
        fullTableName = '%s.%s' %(tableSchema, tableName) if tableSchema else tableName
        sql = "Select OBJECT_ID('%s')" %(fullTableName)
        self.default = sql
        self.connQuery[eConn.SQLSERVER] = sql
        self.connQuery[eConn.LITE] = "SELECT name FROM sqlite_master WHERE type = 'table' AND name = '%s';" %(fullTableName)

    def setSqlDelete (self, sqlFilter, tableName, tableSchema):
        fullTableName = '%s.%s' % (tableSchema, tableName) if tableSchema else tableName
        sql ="Delete From %s where %s " %(fullTableName, sqlFilter)
        self.default = sql
        self.connQuery[eConn.SQLSERVER] = sql

    def tblCopyByColumn(self, tableName, tableSchema, srcTableName, columns):
        sourceTableName = '%s.%s' % (tableSchema, tableName) if tableSchema else srcTableName
        targetTableName = '%s.%s' % (tableSchema, tableName) if tableSchema else tableName
        columns= ",".join(columns)

        sql = "insert into %s (%s) select %s from %s" % (targetTableName,columns,columns, sourceTableName)
        self.default = sql
        self.connQuery[eConn.SQLSERVER] = sql

    def setSqlExistingIndexes(self, tableName):
        sql = """
        SELECT a.name AS indexName, COL_NAME(b.object_id,b.column_id) AS columnName,a.type AS isClustered,a.is_unique As isUnique
        FROM sys.indexes AS a  INNER JOIN sys.index_columns AS b  ON a.object_id = b.object_id AND a.index_id = b.index_id  
        WHERE a.is_hypothetical = 0 AND a.object_id = OBJECT_ID('%s');  """ % (tableName)

        self.default = sql
        self.connQuery[eConn.SQLSERVER] = sql

    def setSqlIndex(self,  tableName, columns, isCluster, isUnique):
        tableIndexName = tableName.split(".")
        tableIndexName = tableIndexName[1] if len(tableIndexName)>1 else tableIndexName[0]
        columnNames = "_".join(columns)
        indexName = 'IX_%s_%s' %(tableIndexName, columnNames)
        isClusterStr = "CLUSTERED" if isCluster else ""
        isUniqueStr  = "UNIQUE" if isUnique else ""

        sql = """CREATE %s %s INDEX %s ON %s (%s)""" %(isUniqueStr,isClusterStr,indexName, tableName, ",".join(columns) )
        self.default = sql
        self.connQuery[eConn.SQLSERVER] = sql


# (c) 2017-2019, Tal Shany <tal.shany@biSkilled.com>
#
# This file is part of dingDong
#
# dingDong is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# dingDong is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with dingDong.  If not, see <http://www.gnu.org/licenses/>.

import time
import logging
import inspect
import os

from dingDong.config import config

### INTERNAL LOGGING CLASSES


class __myLogger (object):
    class __logFilter(object):
        def __init__(self, level):
            self.__level = level

        def filter(self, logRecord):
            return logRecord.levelno <= self.__level

    def __init__ (self, loggLevel=logging.DEBUG, logFormat='%(asctime)s %(levelname)s %(message)s' ):
        dateFormat          = '%Y-%m-%d %H:%M:%S'
        self.logFormatter   = logging.Formatter(logFormat, dateFormat)
        self.logLevel       = loggLevel

        logging.basicConfig(level=self.logLevel, format=logFormat, datefmt=dateFormat)
        self.isLogsFilesInit= False
        self.logTmpFileErr  = None
        self.logTmpFileWar  = None
        self.logg =  logging.getLogger(__name__)

        if config.LOGS_DIR and os.path.isdir(config.LOGS_DIR):
            self.setLogsFiles(logDir=config.LOGS_DIR)

        #if logStdout:
        #    consoleHandler = logging.StreamHandler(sys.stdout)
        #    consoleHandler.setFormatter(self.logFormatter)
        #    self.logg.addHandler(consoleHandler)

    def setLogsFiles (self, logDir=None, timeFormat='%Y%m%d'):
        self.logDir = logDir if logDir else config.LOGS_DIR

        if not self.logDir or len(self.logDir)<1:
            p("DIRECTORY FOR LOGS IS NOT DEFINED ")
            return

        if not os.path.isdir(self.logDir):
            err = "LOGS DIRECTORY: %s IS NOT VALID !" %(self.logDir)
            raise ValueError(err)

        currentDate = time.strftime(timeFormat)
        logFile = config.LOGS_INFO_NAME
        logFile = "%s_%s.log"%(logFile, currentDate) if logFile and ".log" not in logFile.lower() else logFile

        logErrFile = config.LOGS_ERR_NAME
        logErrFile = "%s_%s.err" % (logErrFile, currentDate) if logErrFile and ".err" not in logErrFile.lower() else logErrFile

        logTmpFile = config.LOGS_TMP_NAME
        logTmpFileErr = "%s.err" % (logTmpFile) if logTmpFile and ".err" not in logTmpFile.lower() else logTmpFile
        logTmpFileWar = "%s.warning" % (logTmpFile) if logTmpFile and ".warning" not in logTmpFile.lower() else logTmpFile

        if not os.path.isdir(self.logDir):
            err = "%s if not a correct directory " % self.logDir
            raise ValueError(err)

        if logTmpFile:
            ### Error Temp log file
            self.logTmpFileErr = os.path.join(self.logDir, logTmpFileErr)
            tmpFileErrors   = logging.FileHandler(self.logTmpFileErr, mode='a')
            tmpFileErrors.setFormatter(self.logFormatter)
            tmpFileErrors.setLevel(logging.ERROR)
            tmpFileErrors.addFilter(self.__logFilter(logging.ERROR))
            self.logg.addHandler(tmpFileErrors)

            self.logTmpFileWar = os.path.join(self.logDir, logTmpFileWar)
            tmpFileWarning = logging.FileHandler(self.logTmpFileWar, mode='a')
            tmpFileWarning.setFormatter(self.logFormatter)
            tmpFileWarning.setLevel(logging.WARNING)
            tmpFileWarning.addFilter(self.__logFilter(logging.WARNING))
            self.logg.addHandler(tmpFileWarning)

            try:
                if os.path.isfile(self.logTmpFileErr):
                    open(self.logTmpFileErr, 'w').close()

                if os.path.isfile(self.logTmpFileWar):
                    open(self.logTmpFileWar, 'w').close()
            except:
                p("CANNOT DELETE TEMP FILES\n     %s\n     %s" %(self.logTmpFileErr, self.logTmpFileWar), "e")

        if not logErrFile:
            fileHandler = logging.FileHandler(os.path.join(self.logDir, logFile), mode='a')
            fileHandler.setFormatter(self.logFormatter)
            self.logg.addHandler(fileHandler)
        else:
            # log file info
            fileHandlerInfo = logging.FileHandler(os.path.join(self.logDir, logFile), mode='a')
            fileHandlerInfo.setFormatter(self.logFormatter)
            fileHandlerInfo.setLevel(self.logLevel)
            self.logg.addHandler(fileHandlerInfo)

            # Err file info
            fileHandlerErr = logging.FileHandler(os.path.join(self.logDir, logErrFile), mode='a')
            fileHandlerErr.setFormatter(self.logFormatter)
            fileHandlerErr.setLevel(logging.ERROR)
            self.logg.addHandler(fileHandlerErr)

        ## Delete OLD log file
        self.deleteLogFiles()

    def getLogg (self):
        return self.logg

    def getLogsDir (self):
        return self.logDir

    def setLogLevel (self, logLevel):
        self.logLevel = logLevel
        self.logg.setLevel(self.logLevel)

    def getLogData (self, logPath=None, error=True):
        def getLines (dir=None, f=None, fPath=None):
            lines = None
            logPath = os.path.join(dir, f) if dir and f else fPath

            if logPath and os.path.isfile(logPath):
                with open(logPath) as f:
                    lines = f.read().splitlines()
            return lines

        if logPath:
            lines = getLines (dir=None, f=None, fPath=logPath)
        elif error:
            lines = getLines(dir=self.logDir, f=self.logTmpFileErr, fPath=None)
        else:
            lines = getLines(dir=self.logDir, f=self.logTmpFileWar, fPath=None)

        return lines

    def deleteLogFiles (self, days=None ):
        days = days if days else config.LOGS_HISTORY_DAYS

        if self.logDir:
            now = time.time()
            old = now - (days * 24 * 60 * 60)
            for f in os.listdir(self.logDir):
                path = os.path.join(self.logDir, f)
                if os.path.isfile(path):
                    stat = os.stat(path)
                    if stat.st_mtime < old:
                        self.logg.info("DELETE FILE %s" %(path))
                        os.remove(path)

class __listHandler(logging.Handler):  # Inherit from logging.Handler
    def __init__(self):
        logging.Handler.__init__(self)
        # Our custom argument
        self.log_list = []

    def emit(self, record):
        # record.message is the log message
        self.log_list.append(self.format(record))

    def getList (self):
        return self.log_list


LOGGER_OBJECT  = __myLogger(loggLevel=config.LOGS_DEBUG)
__logg  = LOGGER_OBJECT.getLogg()

def p(msg, ind='I'):
    func = inspect.currentframe().f_back.f_code
    fileName = os.path.split(func.co_filename)[1].replace(".py", "")

    ind = ind.upper()
    indPrint = {'E': 'ERROR>> ',
                'I': 'INFO >> ',
                'W': 'WARNING >>',
                'II': 'DEBUG>> ',
                'III': 'Progress>> '}

    if 'III' in ind:
        __logg.debug("\r %s %s" %(indPrint[ind], msg))
    elif 'II' in ind:
        __logg.debug("%s->%s: %s" %(fileName, func.co_name, msg))
    elif 'I' in ind:
        __logg.info("%s->%s: %s" %(fileName, func.co_name, msg))
    elif 'W' in ind:
        __logg.warning("%s->%s: %s" % (fileName, func.co_name, msg))
    else:
        __logg.error("%s,%s->%s : %s " % (fileName, func.co_firstlineno, func.co_name,msg))


# Copyright (c) 2017-2019, BPMK LTD (BiSkilled) Tal Shany <tal.shany@biSkilled.com>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
#
# This file is part of dingDong
#
# dingDong is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# dingDong is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with dingDong.  If not, see <http://www.gnu.org/licenses/>.

try:
    import queue
except ImportError:
    import Queue as queue

import traceback
import sys
from threading import Thread

from dingDong.bl.ddManager import ddManager

from dingDong.config            import config
from dingDong.misc.logger       import p, LOGGER_OBJECT
from dingDong.bl.jsonParser     import jsonParser
from dingDong.misc.enumsJson    import eJson
from dingDong.conn.baseConnectorManager   import mngConnectors as conn

## Execters
from dingDong.executers.executeSql import execQuery
from dingDong.executers.executeAddMsg import executeAddMsg
from dingDong.executers.executeMicrosoftOLAP import OLAP_Process


class dingDong:
    def __init__ (self,  dicObj=None, filePath=None,
                dirData=None, includeFiles=None, notIncludeFiles=None,
                dirLogs=None,connDict=None, processes=None, sqlFolder=None):

        self.sqlFolder = sqlFolder if sqlFolder else config.SQL_FOLDER_DIR
        self.jsonParser = jsonParser(dicObj=dicObj, filePath=filePath,
                                     dirData=dirData, includeFiles=includeFiles, notIncludeFiles=notIncludeFiles,
                                     connDict=connDict, sqlFolder=self.sqlFolder)


        self.msg = executeAddMsg()
        self.propcesses = processes if processes else config.NUM_OF_PROCESSES

        if dirLogs or config.LOGS_DIR:
            LOGGER_OBJECT.setLogsFiles (logDir=dirLogs)

        ## Defualt properties
        self.connDict       = self.jsonParser.connDict

    def ding (self, destList=None, jsName=None, jsonNodes=None):
        p('STARTING TO MODEL DATA STRUCURE >>>>>' , "i")
        allNodes = self.__getNodes(destList=destList, jsName=jsName, jsonNodes=jsonNodes)

        ## ALL Files
        for jsName, jsonNodes in allNodes:
            ## ALL On all nodes
            for jMap in jsonNodes:
                dingObject = ddManager(node=jMap, connDict=self.connDict)
                dingObject.ding()

        p('FINSHED TO MODEL DATA STRUCURE >>>>>', "i")
        p('>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>', "ii")

    ## There is parrallel processing option
    def dong (self, destList=None, jsName=None, jsonNodes=None):
        p('STARTING TO EXTRACT AND LOAD >>>>>', "i")
        allNodes = self.__getNodes(destList=destList, jsName=jsName, jsonNodes=jsonNodes)
        processList = []

        for jsName, jsonNodes in allNodes:
            procTotal = len(jsonNodes)
            for procNum, jMap in  enumerate (jsonNodes):
                processList.append((jMap, procNum, procTotal))


        numOfProcesses = min (len(processList), self.propcesses)

        if numOfProcesses==1:
            for proc in processList:
                dingObject = ddManager(node=proc[0], connDict=self.connDict)
                dingObject.dong()

        elif numOfProcesses > 1:
            q = queue.Queue(maxsize=0)
            for i, processParams in enumerate (processList):
                q.put(processParams)
                if i%2 ==0 and i>0:
                    for i in range(q.qsize()):
                        worker = Thread(target=self.execDong, args=(q,))
                        worker.setDaemon(True)
                        worker.start()
                    q.join()

            for i in range(q.qsize()):
                worker = Thread(target=self.execDong, args=(q,))
                worker.setDaemon(True)
                worker.start()
            q.join()
        else:
            p("THERE IS NO MODEL TO EXTRACT", "w")
        p('FINISHED TO EXTRACT AND LOAD >>>>>', "i")
        p('>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>', "ii")

    def execDong (self, q):
        while True:
            try:
                (jMap, procNum, procTotal) = q.get()
                dingObject = ddManager(node=jMap)
                if procTotal > 1:
                    p("DONG PROCESS NUMBER %s OUT OF %s" % (str(procNum), str(procTotal)))
                dingObject.dong()
            except Exception as e:
                err = traceback.extract_tb(e.__traceback__)
                p("MULTI THREADING ERROR:\n%s " % e, "e")
                for er in err:
                    p(er , "e")
            finally:
                q.task_done()

    def setLoggingLevel (self, val):
        CRITICAL = 50
        ERROR = 40
        WARNING = 30
        INFO = 20
        DEBUG = 10
        NOTSET = 0

        if val in (CRITICAL, ERROR, WARNING, INFO, DEBUG, NOTSET):
            config.LOGS_DEBUG = val
            LOGGER_OBJECT.logLevel = val
            LOGGER_OBJECT.setLogLevel(logLevel=config.LOGS_DEBUG)
        else:
            err = "Logging is not valid, valid values: 0,10,20,30,40,50"
            raise ValueError(err)

    """ LOADING BUSINESS LOGIC EXECUTOERS """
    def execDbSql(self, queries, connName=None, connType=None, connUrl=None, connPropDic=None):
        connPropDic = connPropDic if connPropDic else {}
        if connName : connPropDic[eJson.jValues.NAME] = connName
        if connType : connPropDic[eJson.jValues.CONN] = connType
        if connUrl  : connPropDic[eJson.jValues.URL] = connUrl

        connObj = conn(connPropDic=connPropDic , connLoadProp=self.connDict)
        execQuery(sqlWithParamList=queries, connObj=connObj)

    def execMicrosoftOLAP (self, serverName, dbName, cubes=[], dims=[], fullProcess=True):
        OLAP_Process(serverName=serverName, dbName=dbName, cubes=cubes, dims=dims, fullProcess=fullProcess)

    def __getNodes(self, destList=None, jsName=None, jsonNodes=None):
        allNodes = []
        if jsName and jsonNodes:
            allNodes = [(jsName, jsonNodes)]
        else:
            for jsonMapping in self.jsonParser.jsonMappings(destList=destList):
                for jsName in jsonMapping:
                    allNodes.append ( (jsName, jsonMapping[jsName]) )
        return allNodes


# (c) 2017-2019, Tal Shany <tal.shany@biSkilled.com>
#
# This file is part of dingDong
#
# dingDong is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# dingDong is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with dingDong.  If not, see <http://www.gnu.org/licenses/>.

import time
import os
import smtplib
from collections            import OrderedDict
from email.mime.multipart   import MIMEMultipart
from email.mime.text        import MIMEText

from dingDong.config        import config
from dingDong.misc.logger   import LOGGER_OBJECT,p
from dingDong.executers.executeHTMLReport import eHtml, createHtmlFromList

class msgProp (object):
    STEP_NUM    = "STEP NUMBER"
    DESC        = "DESCRIPTION"
    TS          = "TIME STAMP"
    STEP_TIME   = "TOTAL TIME CURRENT STEP"
    TOTAL_TIME  = "TOTAL TIME"

    MSG_SUBJECT_SUCCESS = "LOADING JOB %s "
    MSG_SUBJECT_FAILURE = "ERROR LOADING JOB %s "
    MSG_LAST_STEP       = "TOTAL EXECUTION  "

    _TIME_FORMAT = "%m/%d/%Y %H:%M:%S"
    _PREFIX_DESC = "STATE_"

class executeAddMsg (object):

    def __init__ (self, timeFormat=msgProp._TIME_FORMAT, sDesc=msgProp._PREFIX_DESC ):
        self.startTime  = time.time()
        self.lastTime   = self.startTime
        self.stateDic   = OrderedDict()
        self.loggClass  = LOGGER_OBJECT
        self.timeFormat = timeFormat
        self.stateCnt   = 0
        self.sDesc      = sDesc
        self.inProcess  = False

    def addState (self, sDesc=None):
        self.stateCnt+=1
        if not sDesc:
            sDesc="%s%s" %(str(self.sDesc),str(self.stateCnt))
        ts = time.time()
        tsStr = time.strftime(self.timeFormat, time.localtime(ts))
        tCntFromStart   = str(round ( ((ts - self.startTime) / 60) , 2))
        tCntFromLaststep= str(round ( ((ts - self.lastTime) / 60) , 2))
        self.lastTime   = ts
        self.stateDic[self.stateCnt] = OrderedDict({msgProp.STEP_NUM   : self.stateCnt,
                                        msgProp.DESC       :sDesc,
                                        msgProp.TS         :tsStr,
                                        msgProp.STEP_TIME  :tCntFromLaststep,
                                        msgProp.TOTAL_TIME :tCntFromStart })



    def end(self, msg=None,pr=True):
        msg = msg if msg else msgProp.MSG_LAST_STEP
        self.addState(sDesc=msg)

        if pr:
            for col in self.stateDic:
                p (list(self.stateDic[col].values()))

    def sendSMTPmsg (self, msgName, onlyOnErr=False, withErr=True,withWarning=True ):

        okMsg = msgProp.MSG_SUBJECT_SUCCESS %(msgName)
        errMsg= msgProp.MSG_SUBJECT_FAILURE %(msgName)

        errList = self.loggClass.getLogData (error=True)
        errCnt  = len(errList) if errList else 0

        warList = self.loggClass.getLogData (error=False) if withWarning else None

        htmlList = []
        msgSubj  = okMsg if errCnt<1 else errMsg

        if onlyOnErr and errCnt>0 or not onlyOnErr:
            # First table - general knowledge
            self.addState(sDesc='')

            headerNames = None
            dicFirstTable = {eHtml.HEADER:[],eHtml.ROWS:[]}


            for st in self.stateDic:
                if not headerNames:
                    headerNames = list(self.stateDic[st].keys())
                    dicFirstTable[eHtml.HEADER] = headerNames

                dicFirstTable[eHtml.ROWS].append ( list(self.stateDic[st].values()) )

            htmlList.append (dicFirstTable)
            if withErr:
                if errList and len(errList)>0:
                    dicFirstTable = {eHtml.HEADER: ['ERROR FOUNDS IN CURRENT EXECUTION'],eHtml.ROWS: []}
                    for err in errList:
                        dicFirstTable[eHtml.ROWS].append ( [err] )
                else:
                    dicFirstTable = {eHtml.HEADER: ['NO ERROR FOUNDS IN CURRENT EXECUTION'], eHtml.ROWS: []}
                htmlList.append(dicFirstTable)

            if withWarning and warList and len (warList)>0:
                dicFirstTable = {eHtml.HEADER: ['WARNING FOUNDS IN CURRENT EXECUTION'],eHtml.ROWS: []}
                for war in warList:
                    dicFirstTable[eHtml.ROWS].append([war])

                htmlList.append(dicFirstTable)

            msgHtml = createHtmlFromList(htmlList=htmlList, htmlHeader=msgName)
            self.__sendSMTP(msgSubj=msgSubj, msgHtml=msgHtml)

    def __sendSMTP (self, msgSubj, msgHtml=None, msgText=None):
        sender          = config.SMTP_SENDER
        receivers       = ", ".join(config.SMTP_RECEIVERS)
        receiversList   = config.SMTP_RECEIVERS
        serverSMTP      = config.SMTP_SERVER
        serverUsr       = config.SMTP_SERVER_USER
        serverPass      = config.SMTP_SERVER_PASS

        msg = MIMEMultipart('alternative')
        msg['Subject']  = msgSubj
        msg['From']     = sender
        msg['To']       = receivers

        if msgText:
            textInMail = ''
            if isinstance(msgText, list):
                for l in msgText:
                    textInMail += l + "\n"
            else:
                textInMail = msgText

            msg.attach(MIMEText(textInMail, 'plain'))

        if msgHtml and len(msgHtml)>0:
            msg.attach( MIMEText(msgHtml, 'html') )

        try:
            server = smtplib.SMTP(serverSMTP)
            server.ehlo()
            server.starttls()

            server.login(serverUsr, serverPass)
            server.sendmail(sender, receiversList, msg.as_string())
            server.quit()
        except smtplib.SMTPException:
            err = "gFunc->sendMsg: unable to send email to %s, subject is: %s " % (str(receivers), str(msgSubj))
            raise ValueError(err)


# (c) 2017-2019, Tal Shany <tal.shany@biSkilled.com>
#
# This file is part of dingDong
#
# dingDong is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# dingDong is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with dingDong.  If not, see <http://www.gnu.org/licenses/>.

import sys
import shutil
import os
import io
import time
import codecs
from collections import OrderedDict
if sys.version_info[0] == 2:
    import csv23 as csv
else:
    import csv as csv

from dingDong.conn.baseBatch  import baseBatch
from dingDong.conn.baseBatchFunction    import *
from dingDong.misc.enumsJson  import eConn, eJson
from dingDong.misc.misc import uniocdeStr
from dingDong.config          import config
from dingDong.misc.logger     import p

DEFAULTS = {
            eJson.jFile.MIN_SIZE:1024,
            eJson.jFile.DEF_COLUMN_PREF :'col_',
            eJson.jFile.ENCODING:'windows-1255',
            eJson.jFile.DELIMITER:',',
            eJson.jFile.ROW_HEADER:1,
            eJson.jFile.END_OF_LINE:'\r\n',
            eJson.jFile.MAX_LINES_PARSE:50000,
            eJson.jFile.LOAD_WITH_CHAR_ERR:'strict',
            eJson.jFile.APPEND:False,
            eJson.jFile.CSV:False,
            eJson.jFile.REPLACE_TO_NONE:None #r'\"|\t'
           }

DATA_TYPES = {  }

class connFile (baseBatch):

    def __init__ (self, folder=None,fileName=None,
                    fileMinSize=None, colPref=None, encode=None,isCsv=None,
                    delimiter=None,header=None, endOfLine=None,linesToParse=None,
                    withCharErr=None,append=None,replaceToNone=None,
                    isTar=None, isSrc=None, connPropDict=None):


        self.conn = eConn.FILE
        baseBatch.__init__(self, conn=self.conn, connPropDict=connPropDict, defaults=DEFAULTS, dataType=DATA_TYPES)
        self.usingSchema = False

        """ BASIC PROPERTIES FROM BASECONN """
        self.connName       = self.connName

        """ FILE DEFAULTS PROPERTIES """

        self.fileMinSize= self.setProperties (propKey=eJson.jFile.MIN_SIZE, propVal=fileMinSize, propDef=DEFAULTS)
        self.colPref    = self.setProperties(propKey=eJson.jFile.DEF_COLUMN_PREF, propVal=colPref, propDef=DEFAULTS)
        self.encode     = self.setProperties(propKey=eJson.jFile.ENCODING, propVal=encode, propDef=DEFAULTS)
        self.header     = self.setProperties(propKey=eJson.jFile.ROW_HEADER, propVal=header, propDef=DEFAULTS)
        self.maxLinesParse = self.setProperties(propKey=eJson.jFile.MAX_LINES_PARSE, propVal=linesToParse, propDef=DEFAULTS)
        self.withCharErr= self.setProperties(propKey=eJson.jFile.LOAD_WITH_CHAR_ERR, propVal=withCharErr, propDef=DEFAULTS)
        self.delimiter  = self.setProperties(propKey=eJson.jFile.DELIMITER, propVal=delimiter, propDef=DEFAULTS)
        self.endOfLine  = self.setProperties(propKey=eJson.jFile.END_OF_LINE, propVal=endOfLine, propDef=DEFAULTS)
        self.append     = self.setProperties(propKey=eJson.jFile.APPEND, propVal=append, propDef=DEFAULTS)
        self.replaceToNone= self.setProperties(propKey=eJson.jFile.REPLACE_TO_NONE, propVal=replaceToNone, propDef=DEFAULTS)
        self.isCsv      = self.setProperties(propKey=eJson.jFile.CSV, propVal=isCsv, propDef=DEFAULTS)

        """ FILE PROPERTIES """
        self.fileFullName = None
        self.folder   = self.setProperties(propKey=eJson.jValues.FOLDER, propVal=folder)
        if not self.folder:
            self.folder = self.setProperties(propKey=eJson.jValues.URL, propVal=folder)

        self.fileName = self.setProperties(propKey=eJson.jValues.OBJ, propVal=fileName)
        self.isSrc    = self.setProperties(propKey=eJson.jValues.IS_SOURCE, propVal=isSrc)
        self.isTar    = self.setProperties(propKey=eJson.jValues.IS_TARGET, propVal=isTar)

        if self.fileName:
            head, tail = os.path.split (self.fileName)
            if head and len(head)>0 and tail and len (tail)>1:
                self.fileFullName   = self.fileName
                self.folder         = head
            elif self.folder:
                self.fileFullName = os.path.join(self.folder, self.fileName)
            else:
                p(u"THERE IS NO FOLDER MAPPING, FILE CONNENTION FAILED, %s" %(self.fileName), "e")
                return
            if (os.path.isfile(self.fileFullName)):
                p (u"FILE EXISTS:%s, DELIMITER %s, HEADER %s " %(self.fileFullName , self.delimiter ,self.header ), "ii")
        elif self.folder:
            head, tail = os.path.split(self.folder)
            if head and len(head)>0 and tail and len (tail)>1:
                self.folder = head
                self.fileFullName = self.folder

    """  MANADATORY METHOD INHERTIED FROM BASE BATCH INTERFACE"""
    def connect(self, fileName=None):
        if fileName:
            self.fileFullName = fileName
            return True
        elif not self.fileFullName:
            if self.folder and os.path.isdir(self.folder):
                p("CONNETCTED USING FOLDER %s" %self.folder)
                return True
            else:
                err = u"FILE NOT VALID: %s" %(self.fileFullName)
                raise ValueError(err)
        return True

    def close(self):
        pass

    def create(self, stt=None, fullPath=None):
        fullPath = fullPath if fullPath else self.fileFullName
        self.cloneObject(stt=stt, fullPath=fullPath)
        p("NO POINT TO CREATE FILE %s " %(fullPath), "ii")

    """  Strucutre Dictinary for file: {Column Name: {ColumnType:XXXXX} .... }
            Types : STR , FLOAD , INT, DATETIME (only if defined)  """
    def getStructure(self, fullPath=None):
        ret     = OrderedDict()
        fullPath= fullPath if fullPath else self.fileFullName

        if self.isExists(fullPath=fullPath):
            with io.open(fullPath, 'r', encoding=self.encode) as f:
                if not self.header:
                    headers = f.readline().strip(self.endOfLine).split(self.delimiter)
                    if headers and len(headers) > 0:
                        for i, col in enumerate(headers):
                            colStr = '%s%s' %(self.colPref, str(i))
                            ret[colStr] = {eJson.jSttValues.TYPE: self.defDataType, eJson.jSttValues.SOURCE: colStr}
                else:
                    for i, line in enumerate(f):
                        if self.header-1 == i:
                            headers = line.strip(self.endOfLine).split(self.delimiter)
                            if headers and len(headers) > 0:
                                for i, col in enumerate(headers):
                                    ret[col] = {eJson.jSttValues.TYPE: self.defDataType, eJson.jSttValues.SOURCE: col}

                            break
        else:
            p ('FILE NOT EXISTS %s >>> ' %( str(fullPath) ), "ii")

        return ret

    def isExists(self, fullPath=None):
        fullPath = fullPath if fullPath else self.fileFullName
        return os.path.isfile(fullPath)

    def preLoading(self):
        if self.isExists() and self.append:
            p("FILE %s EXISTS WILL APPEND DATA " % (self.fileFullName))

    def extract(self, tar, tarToSrc, batchRows=None, addAsTaret=True):
        batchRows = batchRows if batchRows else self.batchSize
        fnOnRowsDic     = {}
        execOnRowsDic   = {}

        startFromRow    = 0 if not self.header else self.header
        listOfColumnsH  = {}
        listOfColumnsL  = []
        targetColumnList= []

        fileStructure = self.getStructure()
        fileStructureL = OrderedDict()

        for i, col in enumerate(fileStructure):
            fileStructureL[col.lower()] = i
            listOfColumnsH[i] = col

        ## File with header and there is target to source mapping
        if tarToSrc and len(tarToSrc)>0:
            mappingSourceColumnNotExists = u""
            fileSourceColumnNotExists    = u""

            for i, col in enumerate (tarToSrc):
                targetColumnList.append(col)
                if eJson.jSttValues.SOURCE in tarToSrc[col] and tarToSrc[col][eJson.jSttValues.SOURCE]:
                    srcColumnName = tarToSrc[col][eJson.jSttValues.SOURCE]
                    if srcColumnName.lower() in fileStructureL:
                        listOfColumnsL.append(fileStructureL[ srcColumnName.lower() ])
                    else:
                        mappingSourceColumnNotExists+=uniocdeStr (srcColumnName)+u" ; "
                else:
                    listOfColumnsL.append(-1)

                ### ADD FUNCTION
                if eJson.jSttValues.FUNCTION in tarToSrc[col] and tarToSrc[col][eJson.jSttValues.FUNCTION]:
                    fnc = eval(tarToSrc[col][eJson.jSttValues.FUNCTION])
                    fnOnRowsDic[ i ] = fnc if isinstance(fnc, (list, tuple)) else [fnc]

                ### ADD EXECUTION FUNCTIONS
                if eJson.jSttValues.EXECFUNC in tarToSrc[col] and len(tarToSrc[col][eJson.jSttValues.EXECFUNC]) > 0:
                    newExcecFunction = tarToSrc[col][eJson.jSttValues.EXECFUNC]
                    regex = r"(\{.*?\})"
                    matches = re.finditer(regex, tarToSrc[col][eJson.jSttValues.EXECFUNC], re.MULTILINE | re.DOTALL)
                    for matchNum, match in enumerate(matches):
                        for groupNum in range(0, len(match.groups())):
                            colName = match.group(1)
                            if colName and len(colName) > 0 and colName in fileStructureL:
                                colName = colName.replace("{", "").replace("}", "")
                                newExcecFunction.replace(colName, colName)
                    execOnRowsDic[i] = newExcecFunction

            for colNum in listOfColumnsH:
                if colNum not in listOfColumnsL:
                    fileSourceColumnNotExists+= uniocdeStr( listOfColumnsH[colNum] )+u" ; "

            if len(mappingSourceColumnNotExists)>0:
                p("SOURCE COLUMN EXISTS IN SOURCE TO TARGET MAPPING AND NOT FOUND IN SOURCE FILE: %s" %(mappingSourceColumnNotExists),"w")

            if len (fileSourceColumnNotExists)>0:
                p("FILE COLUMN NOT FOUD IN MAPPING: %s" %(fileSourceColumnNotExists),"w")
        ## There is no target to source mapping, load file as is
        else:
            for colNum in listOfColumnsH:
                listOfColumnsL.append(colNum)

        """ EXECUTING LOADING SOURCE FILE DATA """
        rows = []
        try:
            with io.open( self.fileFullName, 'r', encoding=self.encode, errors=self.withCharErr) as textFile:
                if self.isCsv:
                    fFile = csv.reader(textFile, delimiter=self.delimiter)
                    for i, split_line in enumerate(fFile):
                        if i>=startFromRow:
                            if self.replaceToNone:
                                rows.append( [ re.sub(self.replaceToNone,"",split_line[x],re.IGNORECASE|re.MULTILINE|re.UNICODE)  if x>-1 and len(split_line[x])>0 else None for x in listOfColumnsL] )
                            else:
                                rows.append([split_line[x] if x > -1 and len(split_line[x]) > 0 else None for x in listOfColumnsL])

                        if self.maxLinesParse and i>startFromRow and i%self.maxLinesParse == 0:
                            rows = self.dataTransform(data=rows, functionDict=fnOnRowsDic, execDict=execOnRowsDic)
                            tar.load(rows=rows, targetColumn=targetColumnList)
                            rows = list ([])
                else:
                    for i, line in enumerate(textFile):
                        line = re.sub(self.replaceToNone,"",line,re.IGNORECASE|re.MULTILINE|re.UNICODE) if self.replaceToNone else line
                        line = line.strip(self.endOfLine)
                        split_line = line.split(self.delimiter)
                        # Add headers structure
                        if i >= startFromRow:
                            rows.append([split_line[x] if x > -1 and len(split_line[x]) > 0 else None for x in listOfColumnsL])

                        if self.maxLinesParse and i > startFromRow and i % self.maxLinesParse == 0:
                            rows = self.dataTransform(data=rows, functionDict=fnOnRowsDic, execDict=execOnRowsDic)
                            tar.load(rows=rows, targetColumn=targetColumnList)
                            rows = list([])


                if len(rows)>0 : #and split_line:
                    rows = self.dataTransform(data=rows, functionDict=fnOnRowsDic, execDict=execOnRowsDic)
                    tar.load(rows=rows, targetColumn=targetColumnList)
                    rows = list ([])

        except Exception as e:
            p("ERROR LOADING FILE %s  >>>>>>" % (self.fileFullName) , "e")
            p(str(e), "e")

    def load(self, rows, targetColumn):
        totalRows = len(rows) if rows else 0
        if totalRows == 0:
            p("THERE ARE NO ROWS","w")
            return

        if self.append:
            pass

        with codecs.open(filename=self.fileFullName, mode='wb', encoding=self.encode) as f:
            if targetColumn and len(targetColumn) > 0:
                f.write(self.delimiter.join(targetColumn))
                f.write(self.endOfLine)

            for row in rows:
                row = [str(s) for s in row]
                f.write(self.delimiter.join(row))
                f.write(self.endOfLine)

        p('LOAD %s ROWS INTO FILE %s >>>>>> ' % (str(totalRows), self.fileFullName), "ii")
        return

    def execMethod(self):
        pass

    """ PUBLIC METHOD FOR DB MANIPULATION  """

    """ INTERNAL USED for create method
        Create new File is file is exist
        If config.TRACK_HISTORY will save old table as tablename_currentDate   """
    def cloneObject(self, stt=None, fullPath=None):
        fullPath    = fullPath if fullPath else self.fileFullName
        fileName    = os.path.basename(fullPath)
        fileDir     = os.path.dirname(fullPath)

        fileNameNoExtenseion = os.path.splitext(fileName)[0]
        fimeNameExtension = os.path.splitext(fileName)[1]
        ### check if table exists - if exists, create new table
        isFileExists = os.path.isfile(fullPath)
        toUpdateFile = False

        if isFileExists:
            actulSize = os.stat(fullPath).st_size
            if actulSize < self.fileMinSize:
                p("FILE %s EXISTS WITH SIZE SMALLER THAN %s --> WONT UPDATE  ..." %(fullPath, str(actulSize)), "ii")
                toUpdateFile = False

            fileStructure = self.getStructure(fullPath=fullPath)
            fileStructureL= [x.lower() for x in fileStructure]
            sttL = [x.lower() for x in stt]

            if set(fileStructureL) != set(sttL):
                toUpdateFile = True
                p("FILE %s EXISTS, SIZE %s STRUCTURE CHANGED !!" % (fullPath, str(actulSize)), "ii")
            else:
                p("FILE %s EXISTS, SIZE %s STRUCURE DID NOT CHANGED !! " % (fullPath, str(actulSize)), "ii")


            if toUpdateFile and config.TRACK_HISTORY:
                oldName = None
                if (os.path.isfile(fullPath)):
                    oldName = fileNameNoExtenseion + "_" + str(time.strftime('%y%m%d')) + fimeNameExtension
                    oldName = os.path.join(fileDir, oldName)
                    if (os.path.isfile(oldName)):
                        num = 1
                        oldName = os.path.splitext(oldName)[0] + "_" + str(num) + os.path.splitext(oldName)[1]
                        oldName = os.path.join(fileDir, oldName)
                        while (os.path.isfile(oldName)):
                            num += 1
                            FileNoExt = os.path.splitext(oldName)[0]
                            FileExt = os.path.splitext(oldName)[1]
                            oldName = FileNoExt[: FileNoExt.rfind('_')] + "_" + str(num) + FileExt
                            oldName = os.path.join(fileDir, oldName)
                if oldName:
                    p("FILE HISTORY, FILE %s EXISTS, COPY FILE TO %s " % (str(self.fileName), str(oldName)), "ii")
                    shutil.copy(fullPath, oldName)

    """ INTERNAL METHOD  """
    def merge (self, fileName, hearKeys=None, sourceFile=None):
        raise NotImplementedError("count rows need to be implemented")


# -*- coding: utf-8 -*-
# (c) 2017-2019, Tal Shany <tal.shany@biSkilled.com>
#
# This file is part of dingDong
#
# dingDong is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# dingDong is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with dingDong.  If not, see <http://www.gnu.org/licenses/>.

from collections import OrderedDict

from dingDong.conn.connGlobalDB   import baseGlobalDb
from dingDong.misc.enumsJson      import eConn, eJson, findProp

DEFAULTS    = { eJson.jValues.DEFAULT_TYPE: 'varchar(100)', eJson.jValues.SCHEMA: 'dbo',
                eJson.jValues.EMPTY: 'Null', eJson.jValues.COLFRAME: ("[", "]"), eJson.jValues.SP: {}}

DATA_TYPES  = { eConn.dataType.DB_VARCHAR:['varchar', 'longchar', 'bit', 'ntext'],
                    eConn.dataType.DB_INT:['integer', 'counter'],
                    eConn.dataType.DB_FLOAT:['double'],
                    eConn.dataType.DB_DECIMAL:['decimal']
                    }


class access (baseGlobalDb):
    def __init__ (self, conn=None, connUrl=None, connExtraUrl=None, connName=None,
                  connObj=None, connPropDict=None, connFilter=None, connIsTar=False,
                  connIsSrc=False, connIsSql=False):

        baseGlobalDb.__init__(self, conn=conn, connUrl=connUrl, connExtraUrl=connExtraUrl, connName=connName,
                                    connObj=connObj, connPropDict=connPropDict, connFilter=connFilter,
                                    connIsTar=connIsTar, connIsSrc=connIsSrc, connIsSql=connIsSql)

    def getStructure(self, tableName=None, tableSchema=None, sqlQuery=None):
        return self.__getAccessStructure(tableSchema=tableSchema, tableName=tableName)


    """ INTERNAL METHOD  """
    """ INTERNAL USE for getStrucure methid return ACCESS Structure dictionary  """
    def __getAccessStructure (self,tableSchema, tableName):
        tableName = '%s.%s' %(tableSchema, tableName) if tableSchema else tableName
        ret = OrderedDict()

        if self.conn == eConn.ACCESS:
            for row in self.cursor.columns():
                val = {eJson.jSttValues.TYPE:self.defDataType, eJson.jSttValues.ALIACE:None}

                if len(row) > 3:
                    curTblName = row[2].encode("utf-8")
                    if curTblName == tableName:
                        colName = row.column_name.encode("utf-8")
                        colType = row.type_name.lower()
                        if colType in DATA_TYPES[eConn.ACCESS][eConn.dataType.DB_VARCHAR]:
                            colDefType = 'varchar(MAX)' if row.column_size > 4098 else 'varchar(%s)' % str(row.column_size)
                        elif colType in DATA_TYPES[eConn.ACCESS][eConn.dataType.DB_DECIMAL]:
                            colDefType = 'decimal(%s,%s)' % (str(row.column_size), str(row.decimal_digits))
                        else:
                            colDefType = colType

                        val[eJson.jSttValues.TYPE] = colDefType

                        ret[colName] = val
        return ret


